# Fitter functions using curve_fit as the backend
# Derek Fujimoto
# Nov 2020

from bfit.fitting.fit_bdata import fit_bdata
from bfit.fitting.fitter import fitter as fit_base

class fitter(fit_base):
    
    __name__ = 'curve_fit (trf)'
    
    def _do_fit(self, data, fn, omit=None, rebin=None, shared=None, hist_select='', 
                xlims=None, asym_mode='c', fixed=None, parnames=None, **kwargs):
        """Inputs match fit_bdata"""
        return fit_bdata(data, 
                         fn, 
                         omit=omit, 
                         rebin=rebin, 
                         shared=shared, 
                         hist_select=hist_select, 
                         xlims=xlims, 
                         asym_mode=asym_mode, 
                         fixed=fixed, 
                         minimizer='trf', 
                         **kwargs)
            



# Model the fit results with a function
# Derek Fujimoto
# Nov 2019


from tkinter import *
from tkinter import ttk
from functools import partial

import logging, re, os, warnings
import numpy as np
import pandas as pd
import bdata as bd

from bfit.backend.ConstrainedFunction import ConstrainedFunction as CstrFnGenerator
from bfit.fitting.fit_bdata import fit_bdata
from bfit.gui.template_fit_popup import template_fit_popup
from bfit.gui.popup_ongoing_process import popup_ongoing_process 

from multiprocessing import Process, Queue
import queue

# ========================================================================== #
class popup_fit_constraints(template_fit_popup):
    """
        Popup window for modelling the fit results with a function
        
        bfit
        fittab
        logger
        
        output_par_text     text, detected parameter names
        output_text         dict, keys: p0, blo, bhi, res, err, value: tkk.Text objects
       
        output_par_text_val string, contents of output_par_text
        output_text_val     dict of strings, contents of output_text
       
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}
    
    window_title = 'Fit data with contrained parameters'
    
    # ====================================================================== #
    def __init__(self, bfit, constr_text='', output_par_text='', output_text=''):
        
        super().__init__(bfit, constr_text, output_par_text, output_text)
        
        # Keyword parameters
        key_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = CstrFnGenerator.keyvars
        
        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame, text=s, justify=LEFT)
        
        # fit parameter names 
        fit_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved function parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(), 
                                        self.fittab.n_component.get())
        
        s += '\n'.join([k for k in sorted(self.parnames)]) 
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame, text=s, justify=LEFT)

        # module names 
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'
        
        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)
        
        # Text entry
        self.entry_label['text'] = 'Enter one constraint equation per line.'+\
                                 '\nNon-reserved words are shared variables.'+\
                                 '\nEx: "1_T1 = a*np.exp(b*BIAS**0.5)+c"'+\
                                 '\nNote: Shared and fixed flags from main window ignored.'
                
        # gridding
        key_param_label.grid(column=0, row=0)
        fit_param_label.grid(column=0, row=0)
        modules_label.grid(column=0, row=0)
        
        key_param_frame.grid(column=0, row=0, rowspan=1, sticky=(E, W), padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky=(E, W), padx=1, pady=1, rowspan=2)
        fit_param_frame.grid(column=0, row=3, sticky=(E, W, N, S), padx=1, pady=1)
        
    # ====================================================================== #
    def _do_fit(self, text):
        """
            Set up the fit functions and do the fit. Then map the outputs to the
            proper displays. 
        """
        
        self.logger.info('Starting fit')
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # check that the defined variables all match function inputs
        for d in defined: 
            if d not in self.parnames:
                errmsg = 'Definition for "%s" invalid. ' % d+\
                         'Must only define function inputs. '
                messagebox.showerror("Error", errmsg)
                raise RuntimeError(errmsg)
        
        # make shared parameters for the rest of the parameters
        allpar = self.new_par['name'].tolist()
        alldef = defined[:]     # all parameter names in order
        sharelist = [True]*len(allpar)
        
        for n in sorted(self.parnames):
            if n not in defined:
                eqn.append(n)
                alldef.append(n)
                allpar.append(n)
                sharelist.append(False)
                        
        # replace 1_T1 with lambda1
        for i, _ in enumerate(allpar):
            if '1_T1' in allpar[i]:
                allpar[i] = allpar[i].replace('1_T1', 'lambda1')
        
        for i, _ in enumerate(eqn):
            while '1_T1' in eqn[i]:
                eqn[i] = eqn[i].replace('1_T1', 'lambda1')
                
        # make constrained functions
        cgen= CstrFnGenerator(alldef, eqn, allpar, self.parnames)
        
        # get the functions and initial parameters
        fit_files = self.bfit.fit_files
        fetch_files = self.bfit.fetch_files
        fitfns = []
        par = []
        rebin = []
        omit = []
        fnptrs = []
        constr_fns = []
        
        keylist = sorted(fit_files.fit_lines.keys())
        for k in keylist:
            line = fetch_files.data_lines[k]
            data = line.bdfit
            
            # get pulse length
            pulse_len = -1
            try:
                pulse_len = data.bd.pulse_s
            except KeyError:
                pass
            
            # get function
            fn = fit_files.fitter.get_fn(fn_name=fit_files.fit_function_title.get(), 
                                         ncomp=fit_files.n_component.get(), 
                                         pulse_len=pulse_len, 
                                         lifetime=bd.life[fit_files.probe_label['text']])
            
            genf, genc = cgen(data=data, fn=fn)
            fitfns.append(genf)
            fnptrs.append(fn)
            constr_fns.append(genc)
            
            # get initial parameters
            par.append(data.fitpar)
            
            # get rebin
            rebin.append(data.rebin.get())
            
            # get bin omission
            omit.append(data.omit.get())
        
        # clean up omit strings
        for i, om in enumerate(omit):
            if om == fetch_files.bin_remove_starter_line:
                omit[i] = ''
        
        # set up p0, bounds
        p0 = self.new_par['p0'].values
        blo = self.new_par['blo'].values
        bhi = self.new_par['bhi'].values
        
        p0 = [[p]*len(keylist) for p in p0]
        blo = [[p]*len(keylist) for p in blo]
        bhi = [[p]*len(keylist) for p in bhi]
                
        for n in sorted(self.parnames):
            if n not in defined:
                p0.append( [p['p0' ][n] for p in par])
                blo.append([p['blo'][n] for p in par])
                bhi.append([p['bhi'][n] for p in par])
        
        p0 = np.array(p0).T
        blo = np.array(blo).T
        bhi = np.array(bhi).T
        
        # set up fitter inputs
        npar = len(sharelist)
        bounds = [[l, h] for l, h in zip(blo, bhi)]
        data = [self.bfit.data[k] for k in keylist]
        kwargs = {'p0':p0, 'bounds':bounds}
        
        # get minimizer
        if 'trf'   in fit_files.fitter.__name__:  minimizer = 'trf'
        if 'minos' in fit_files.fitter.__name__:  minimizer = 'minos'
        if 'hesse' in fit_files.fitter.__name__:  minimizer = 'migrad'
        
        # set up queue for results
        que = Queue()
        
        # do fit
        def run_fit():
            try:
                out = fit_bdata(data=data, 
                                fn=fitfns, 
                                shared=sharelist, 
                                asym_mode='c', 
                                rebin=rebin, 
                                omit=omit, 
                                xlims=None, 
                                hist_select=self.bfit.hist_select, 
                                minimizer=minimizer, 
                                **kwargs)
            except Exception as err:
                que.put(str(err))
                raise err from None
                
            # par, std_l, std_u, cov, chi, gchi
            que.put(out)
            
        # start the fit
        def do_enable():
            fit_files.input_enable_disable(self.win, state='normal', first=False)
            fit_files.input_enable_disable(fit_files.fit_data_tab, state='normal')
        def do_disable():
            fit_files.input_enable_disable(self.win, state='disabled', first=False)
            fit_files.input_enable_disable(fit_files.fit_data_tab, state='disabled')
            
        popup = popup_ongoing_process(self.bfit, 
                    target = run_fit,
                    message="Constrained fit in progress...", 
                    queue = que,
                    do_disable = do_disable,
                    do_enable = do_enable,
                    )
            
        output = popup.run()
        
        # fit success
        if type(output) is tuple:
            par, std_l, std_u, cov, chi, gchi = output
            std_l = np.abs(std_l)
        
        # error
        elif type(output) is str:
            messagebox.showerror("Error", output)
            return 
        
        # fit cancelled
        elif output is None:
            return
            
        # calculate original parameter equivalents
        for i, k in enumerate(keylist):
            data = fetch_files.data_lines[k].bdfit

            # calculate parameter values and estimate errors
            old_par = [cfn(*par[i]) for cfn in constr_fns[i]]
            old_std_l = [abs(p-cfn(*(par[i]-std_l[i]))) for p, cfn in zip(old_par, constr_fns[i])]
            old_std_u = [abs(p-cfn(*(par[i]+std_u[i]))) for p, cfn in zip(old_par, constr_fns[i])]
            
            old_chi = chi[i]
            
            # set to fitdata containers
            results = pd.DataFrame({'res': old_par, 
                                    'dres+': old_std_u,
                                    'dres-': old_std_l,
                                    'chi': old_chi,
                                    }, index=cgen.oldpar)
            data.set_fitresult({'fn': fnptrs[i], 'results': results, 'gchi': gchi})
            
        # display in fit_files tab
        for key in fit_files.fit_lines:
            fit_files.fit_lines[key].show_fit_result()
        
        # show global chi
        fit_files.gchi_label['text'] = str(np.around(gchi, 2))

        # do end-of-fit stuff
        fit_files.do_end_of_fit()
        
        self.logger.info('Fitting end')
        
        return (par[0, :], std_l[0, :], std_u[0, :])



import sys

# functions
# ~ import bfit.test.functions as fn
# ~ fn.test_lorentzian()
# ~ fn.test_bilorentzian()
# ~ fn.test_gaussian()
# ~ fn.test_quadlorentzian()
# ~ fn.test_pulsed_exp()
# ~ fn.test_pulsed_strexp()

# numeric integration
# ~ import bfit.test.numeric_integration as ni
# ~ ni.test_numeric_integration()

# least squares
# ~ import bfit.test.leastsquares as ls
# ~ ls.test_no_errors()
# ~ ls.test_dy()
# ~ ls.test_dx()
# ~ ls.test_dxdy()
# ~ ls.test_dya()
# ~ ls.test_dxa()
# ~ ls.test_dx_dya()
# ~ ls.test_dxa_dy()
# ~ ls.test_dxa_dya()

# minuit
# ~ import bfit.test.minuit as mnt
# ~ mnt.test_start()
# ~ mnt.test_name()
# ~ mnt.test_error()
# ~ mnt.test_limit()
# ~ mnt.test_fix()

# global fitting
# ~ import bfit.test.global_fitter as gf
# ~ gf.test_constructor()
# ~ gf.test_fitting()

# inspect tab
# ~ import bfit.test.tab_fileviewer as tfview
# ~ tfview.test_fetch(40123, 2020, '20')
# ~ tfview.test_fetch(40033, 2020, '1f')
# ~ tfview.test_fetch(40037, 2020, '1w')
# ~ tfview.test_fetch(40011, 2020, '1n')
# ~ tfview.test_fetch(45539, 2019, '2h')
# ~ tfview.test_fetch(40326, 2019, '2e')
# ~ tfview.test_draw(40123, 2020, '20')
# ~ tfview.test_draw(40033, 2020, '1f')
# ~ tfview.test_draw(40037, 2020, '1w')
# ~ tfview.test_draw(40011, 2020, '1n')
# ~ tfview.test_draw(45539, 2019, '2h')
# ~ tfview.test_draw(40326, 2019, '2e')
# ~ tfview.test_autocomplete()
# ~ tfview.test_draw_mode()
# ~ b.do_close_all()

# fetch tab
# ~ import bfit.test.tab_fetch_files as tfetch
# ~ tfetch.test_fetch()
# ~ tfetch.test_remove()
# ~ tfetch.test_checkbox()
# ~ tfetch.test_draw()

# fit tab
import bfit.test.tab_fit_files as tfit
# ~ tfit.test_populate()
# ~ tfit.test_populate_param()
tfit.test_fit(tfit.separate_curve_fit, 'curve_fit')
# ~ tfit.test_fit(tfit.separate_migrad, 'migrad_hesse')
# ~ tfit.test_fit(tfit.separate_minos, 'migrad_minos')
tfit.test_fit_single(tfit.separate_curve_fit, 'curve_fit')
# ~ tfit.test_fit_single(tfit.separate_migrad, 'migrad_hesse')
# ~ tfit.test_fit_single(tfit.separate_minos, 'migrad_minos')
# ~ tfit.test_fixed()
# ~ tfit.test_shared()
# ~ tfit.test_modify_for_all_reset_p0()
# ~ tfit.test_fit_input()
# ~ tfit.test_p0_prior()
# ~ tfit.test_result_as_p0()
############### ~ tfit.test_draw_fit_results()

# calculator nmr rf attenuation
# ~ import bfit.test.calculator_nmr_atten as calc_nmr_atten
# ~ calc_nmr_atten.test_calc()

# calculator nmr B1
# ~ import bfit.test.calculator_nmr_B1 as calc_nmr_b1
# ~ calc_nmr_b1.test_calc()

# calculator nqr B0
# ~ import bfit.test.calculator_nqr_B0 as calc_nqr_b0
# ~ calc_nqr_b0.test_calc()





# Calculate needed Vpp from desired H1 magnetic field. 
# Derek Fujimoto
# December 2017

from tkinter import *
from tkinter import ttk
from bfit import logger_name
import numpy as np
import webbrowser
import logging
import bfit

# =========================================================================== #
class calculator_nmr_B1(object):
    
    # ======================================================================= #
    def __init__(self, commandline=False):
        """Draw window for calculating needed Vpp from desired H1 magnetic field."""
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # root 
        root = Toplevel()
        root.title("Gerald's Caculator")
        root.columnconfigure(0, weight=1)
        root.rowconfigure(0, weight=1)
        
        # icon
        try:
            img = PhotoImage(file=bfit.icon_path)
            root.tk.call('wm', 'iconphoto', root._w, img)
        except Exception as err:
            print(err)
        
        # variables
        self.field = StringVar()
        self.field.set("")
        self.volt = StringVar()
        self.volt.set("")
        self.nu = StringVar()
        
        self.nu.set('41.27') # MHz
        
        # main frame
        mainframe = ttk.Frame(root, pad=5)
        mainframe.grid(column=0, row=0, sticky=(N, W))
        mainframe.columnconfigure(0, weight=1)
        mainframe.rowconfigure(0, weight=1)
        
        # Entry and other objects
        title_line = ttk.Label(mainframe,   
                text='BNMR Magnetic Oscillating Field -- '+\
                     'Antenna Voltage (peak-peak)', 
                     justify=CENTER)
        self.entry_field = Entry(mainframe, textvariable=self.field, width=10, 
                justify=RIGHT)
        gauss = ttk.Label(mainframe, text='Gauss')
        equals = ttk.Label(mainframe, text='=')
        self.entry_voltage = Entry(mainframe, textvariable=self.volt, 
                width=10, justify=RIGHT)
        voltage = ttk.Label(mainframe, text='millivolts')
        
        # frequency input
        freq_frame = ttk.Frame(root, pad=5)
        
        freq1 = ttk.Label(freq_frame, text='Freq:')
        self.entry_freq = Entry(freq_frame, textvariable=self.nu, width=10, 
                                    justify=RIGHT)
        freq2 = ttk.Label(freq_frame, text='MHz')
        
        freq1.grid(column=0, row=0, padx=5)
        self.entry_freq.grid(column=1, row=0, padx=5)
        freq2.grid(column=2, row=0, padx=5)
        
        # caveat
        warning = ttk.Label(root, 
                            text='WARNING: Calculation only valid for runs \n'+\
                                 'with silver antenna (before Sept 2019)', 
                            justify=CENTER)
        
        # Gridding
        title_line.grid(        column=0, row=0, padx=5, pady=5, columnspan=5)
        freq_frame.grid(        column=0, row=1, padx=5, pady=5, columnspan=5)
        self.entry_field.grid(  column=0, row=2, padx=5, pady=5, sticky=E)
        gauss.grid(             column=1, row=2, padx=5, pady=5, sticky=E)
        equals.grid(            column=2, row=2, padx=20, pady=5, sticky=(E, W))
        self.entry_voltage.grid(column=3, row=2, padx=5, pady=5, sticky=W)
        voltage.grid(           column=4, row=2, padx=5, pady=5, sticky=W)
        warning.grid(           column=0, row=2, padx=5, pady=5, columnspan=5)
        
        self.root = root
        
        # tie key release to calculate 
        self.entry_field.bind('<KeyRelease>', self.calculate)
        self.entry_voltage.bind('<KeyRelease>', self.calculate)
        
        # mainloop
        if commandline:
            root.update()
            root.update_idletasks()
        else:
            self.logger.debug('Initialization success. Starting mainloop.')
            root.mainloop()
        
    # ======================================================================= #
    def calculate(self, *args):
        
        # check focus
        focus_id = str(self.root.focus_get())
        
        # get freqeuency
        nu = float(self.nu.get())
        
        # convert field to voltage
        if focus_id == str(self.entry_field):        
            try:
                field = float(self.field.get())
                value = field/0.0396*nu
                self.volt.set("%9.6f" % np.around(value, 6))
                
                self.logger.debug('Field of %g G converted to voltage of %g mV', 
                                 field, value)
            except ValueError:
                self.volt.set('')
        
        # convert voltage to field
        elif focus_id == str(self.entry_voltage):        
            try:
                voltage = float(self.volt.get())
                value = voltage*0.0396/nu
                self.field.set("%9.6f" % np.around(value, 6))
                
                self.logger.debug('Voltage of %g mV converted to field of %g G', 
                                 voltage, value)
            except ValueError:
                self.field.set('')
            





# Model the fit results with a function
# Derek Fujimoto
# Nov 2019


from tkinter import *
from tkinter import ttk
from functools import partial

import logging, re, os, warnings
import numpy as np
import pandas as pd
import bdata as bd
import weakref as wref

from bfit import logger_name
from bfit.backend.entry_color_set import on_focusout, on_entry_click
from bfit.backend.raise_window import raise_window
import bfit.backend.colors as colors

from bfit.fitting.fit_bdata import fit_bdata
from bfit.backend.ParameterFunction import ParameterFunction as ParFnGenerator

# ========================================================================== #
class popup_add_param(object):
    """
        Popup window for adding parameters for fitting or drawing
        
        bfit
        fittab
        logger
        
        input_fn_text       string: input lines
        
        new_par:            dict: {parname, string: par equation}
        set_par:            dict: {parname, fn handle: lambda : return new_par}
        parnames:           list, function inputs
        reserved_pars:      dict, define values in bdata that can be accessed
        win:                Toplevel
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}
    
    window_title = 'Add parameter for fitting or drawing'
    
    # ====================================================================== #
    def __init__(self, bfit, input_fn_text=''):
        
        self.bfit = bfit
        self.fittab = bfit.fit_files
        
        self.input_fn_text = input_fn_text
        self.set_par = {}
        self.new_par = {}
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.info('Initializing')
        
        # make a new window
        self.win = Toplevel(bfit.mainframe)
        self.win.title(self.window_title)
        frame = ttk.Frame(self.win, relief='sunken', pad=5)
        left_frame = ttk.Frame(frame)
        right_frame = ttk.Frame(frame)

        # set icon
        self.bfit.set_icon(self.win)

        # Key bindings
        self.win.bind('<Control-Key-Return>', self.do_add)
        self.win.bind('<Control-Key-KP_Enter>', self.do_add)
        
        # Text entry
        entry_frame = ttk.Frame(right_frame, relief='sunken', pad=5)
        self.entry_label = ttk.Label(entry_frame, justify=LEFT, text='')
        self.entry = Text(entry_frame, width=60, height=13, state='normal')
        self.entry.bind('<KeyRelease>', self.get_input)
        scrollb = Scrollbar(entry_frame, command=self.entry.yview)
        self.entry['yscrollcommand'] = scrollb.set
        
        # Insert default text
        self.entry.insert('1.0', self.input_fn_text.strip())
        
        # add button 
        add_button = ttk.Button(right_frame, text='Set Parameters', command=self.do_add)
        
        # gridding
        self.entry_label.grid(column=0, row=0, sticky=W)
        self.entry.grid(column=0, row=1)
        
        # grid to frame
        frame.grid(column=0, row=0)
        left_frame.grid(column=0, row=0, sticky=(N, S))
        right_frame.grid(column=1, row=0, sticky=(N, S))
        
        entry_frame.grid(column=0, row=0, sticky=(N, E, W), padx=1, pady=1)
        add_button.grid(column=0, row=1, sticky=(N, E, W, S), padx=20, pady=20)
        right_frame.rowconfigure(1, weight=1)
        
        # initialize 
        self.left_frame = left_frame
        
        # Keyword parameters
        key_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved variable names:\n\n'
        self.reserved_pars = ParFnGenerator.keyvars
        
        keys = list(self.reserved_pars.keys())
        descr = [self.reserved_pars[k] for k in self.reserved_pars]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        key_param_label = ttk.Label(key_param_frame, text=s, justify=LEFT)
        
        # fit parameter names 
        fit_param_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved parameter names:\n\n'
        self.parnames = self.fittab.fitter.gen_param_names(
                                        self.fittab.fit_function_title.get(), 
                                        self.fittab.n_component.get())
        
        s += '\n'.join([k for k in sorted(self.parnames)]) 
        s += '\n'
        fit_param_label = ttk.Label(fit_param_frame, text=s, justify=LEFT)

        # module names 
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'
        
        keys = list(self.modules.keys())
        descr = [self.modules[k] for k in self.modules]
        maxk = max(list(map(len, keys)))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in zip(keys, descr)])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)
        
        # Text entry
        self.entry_label['text'] = 'Enter one parameter equation per line.'+\
                '\nLHS must use only reserved words, constants, or functions'+\
                '\nfrom the reserved modules in the parameter definition.'+\
                '\nEx: "mypar = 1/(1_T1*TEMP)"'+\
                '\n\nAccepts LaTeX input for the new parameter.'+\
                '\nEx: "$\eta_\mathrm{f}$ (Tesla) = B0 * np.exp(-amp**2)"' +\
                '\n\nValues taken as shown in fit results, with no unit scaling.'
                
        # gridding
        key_param_label.grid(column=0, row=0)
        fit_param_label.grid(column=0, row=0)
        modules_label.grid(column=0, row=0)
        
        key_param_frame.grid(column=0, row=0, rowspan=1, sticky=(E, W), padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky=(E, W), padx=1, pady=1, rowspan=2)
        fit_param_frame.grid(column=0, row=3, sticky=(E, W, N, S), padx=1, pady=1)
        
        self.logger.debug('Initialization success. Starting mainloop.')
        
    # ====================================================================== #
    def cancel(self):
        self.win.destroy()
        
    # ====================================================================== #
    def do_add(self, *args):
        """
            Add the parameter and the corresponding function
        """
        # reset draw comp
        self.bfit.fit_files.draw_components = list(self.bfit.draw_components)
        
        # set the parameters
        self.set_par = {}
        try:
            self.set_par = {p: ParFnGenerator(p, e, self.parnames, self.bfit) \
                        for p, e in self.new_par.items()}
        except SyntaxError: # on empty set
            pass
            
        self.logger.info('Added new parameters ', self.set_par)
        
        # update the lists
        self.bfit.fit_files.populate_param()
        
    # ====================================================================== #
    def do_parse(self, *args):
        """
            Detect new global variables
            returns split lines, new parameter names 
        """
        
        # clean input
        text = self.input_fn_text.split('\n')
        text = [t.strip() for t in text if '=' in t]
        
        # check for no input
        if not text:
            self.new_par = {}
            return
        
        # get equations and defined variables
        defined = [t.split('=')[0].strip() for t in text]
        eqn = [t.split('=')[1].strip() for t in text]
        
        # set fields
        self.new_par = {k:e for k, e in zip(defined, eqn)}
        
        # save input
        self.input_fn_text = '\n'.join(text)
        
        # logging
        self.logger.info('Parse found parameters %s', self.new_par)
    
    # ====================================================================== #
    def get_input(self, *args):
        """Get input from text box."""
        self.input_fn_text = self.entry.get('1.0', END)
        self.do_parse()
        
    # ====================================================================== #
    def yview(self, *args):
        """
            Scrollbar for all output text fields
        """
        self.output_par_text.yview(*args)
        for k in self.output_text:
            self.output_text[k].yview(*args)

    



#!/usr/bin/python3

# Test of the numeric integration used when calculating the pulsed stretched
# exponential function. Specifically, check how well the result converges to the
# analytic solution when the stretching exponent beta = 1.

# Ryan M. L. McFadden
# 2021-02-22


import numpy as np
import pandas as pd
import bdata as bd
from bfit.fitting.functions import pulsed_strexp, pulsed_exp


# tolerance = upper limit for an acceptable absolute deviation between results
# n_samples = number of points to use in each array in the simulation
# print_summary = print a summary of simulation results to stdout
def test_numeric_integration(tolerance=1e-5, n_samples=100, print_summary=False):

    # constants appropriate for most β-NMR data taken at TRIUMF
    nuclear_lifetime = bd.life["Li8"]
    pulse_duration = 3.0 * nuclear_lifetime

    # create the SLR functions
    fcn_exp = pulsed_exp(nuclear_lifetime, pulse_duration)
    fcn_strexp = pulsed_strexp(nuclear_lifetime, pulse_duration)

    # generate random values for the SLR fit function parameters that are uniformly
    # distributed between typical bounds for real data
    relaxation_rates = np.random.uniform(
        1e-2 * nuclear_lifetime, 1e2 * nuclear_lifetime, n_samples
    )
    amplitudes = np.random.uniform(0.00, 0.15, n_samples)

    # stretching exponent
    beta = 1.0

    # machine precision for floating point values
    epsilon = np.finfo(float).eps

    # create an empty DataFrame to hold all of the results
    df = pd.DataFrame(
        columns=[
            "Time (s)",
            "Amplitude",
            "Rate (1/s)",
            "Difference",
            "Absolute Difference",
            "Tolerance",
            "Tolerance Exceeded",
            "Epsilon",
            "Epsilon Exceeded",
        ]
    )

    # loop over the random function parameters
    for rate, amplitude in zip(relaxation_rates, amplitudes):

        # generate series of random times to evaluate the SLR fit functions that are
        # uniformly distributed between typical bounds for real data
        time = np.random.uniform(0.0, pulse_duration + 10 * nuclear_lifetime, n_samples)

        # evaluate the difference
        difference = fcn_exp(time, rate, amplitude) - fcn_strexp(
            time, rate, beta, amplitude
        )

        # fill the DataFrame row-by-row
        # https://stackoverflow.com/a/42837693
        for t, d in zip(time, difference):
            df = df.append(
                {
                    "Time (s)": t,
                    "Amplitude": amplitude,
                    "Rate (1/s)": rate,
                    "Difference": d,
                    "Absolute Difference": np.abs(d),
                    "Tolerance": tolerance,
                    "Tolerance Exceeded": np.abs(d) > tolerance,
                    "Epsilon": epsilon,
                    "Epsilon Exceeded": np.abs(d) < epsilon,
                },
                ignore_index=True,
            )

    # optionally print a summary of the results
    if print_summary:
        print("")
        print("-------")
        print("Summary of `test_numeric_integration()`")
        print("-------")
        print("")
        print("Tolerance        = %g" % tolerance)
        print("Machine Epsilon  = %g" % epsilon)
        print("")
        print("Absolute Difference:")
        print(" - Max  = %g" % df["Absolute Difference"].max())
        print(" - Min  = %g" % df["Absolute Difference"].min())
        print(" - Mean = %g" % df["Absolute Difference"].mean())
        print("")
        print("Column Data:")
        print("")
        print(df["Tolerance Exceeded"].value_counts())
        print("")
        print(df["Epsilon Exceeded"].value_counts())
        print("")

    # finally, check all values and raise an error if the tolerance is exceeded
    for i in df.index:
        if df["Tolerance Exceeded"][i]:
            raise AssertionError(
                "Absolute difference greater than tolerance!\n\n"
                + "\t|%.4e| > %.4e\n\n" % (df["Difference"][i], tolerance)
                + "\ttime      = %.4e s\n" % df["Time (s)"][i]
                + "\tamplitude = %.4e\n" % df["Amplitude"][i]
                + "\trate      = %.4e 1/s" % df["Rate (1/s)"][i]
            )



# test inspect tab
# Derek Fujimoto
# Feb 2021

from bfit.test.testing import *
import numpy as np
import matplotlib.pyplot as plt

# get bfit object and tab
tab = b.fileviewer

def test_fetch(r, y, mode):    
    tab.year.set(y)
    tab.runn.set(r)
    test_action(tab.get_data, "fileviewer fetch %s (%d.%d) data" % (mode, y, r))
    test_perfect(tab.data.run, r, "fileviewer fetch %s (%d.%d) data accuracy" % (mode, y, r))
    
def test_draw(r, y, mode):
    
    # get data
    tab.year.set(y)
    tab.runn.set(r)
    tab.get_data()
    
    # draw
    n = len(tab.entry_asym_type['values'])
    
    for i in range(n):
        
        # switch draw types
        tab.entry_asym_type.current(i)
        draw_type = tab.asym_type.get()
        
        # draw
        test_action(tab.draw, "fileviewer draw %s in mode %s" % (mode, draw_type), 'inspect')
        
        if mode == '2e':
            b.do_close_all()
    
def test_draw_mode():
    
    tab.year.set(2020)
    
    # test stack
    b.draw_style.set('stack')
    
    tab.runn.set(40123)
    tab.get_data()
    tab.entry_asym_type.current(0) # set combined asym mode
    
    tab.draw('inspect')
    
    tab.runn.set(40127)
    tab.get_data()
    tab.draw('inspect')
    
    ax = plt.gca()
    test_perfect(len(ax.draw_objs), 2, 'fileviewer stack')
    
    # test redraw
    b.draw_style.set('redraw')
    
    tab.runn.set(40123)
    tab.get_data()
    tab.entry_asym_type.current(0) # set combined asym mode
    
    tab.draw('inspect')
    
    tab.runn.set(40127)
    tab.get_data()
    tab.draw('inspect')
    
    ax = plt.gca()
    test_perfect(len(ax.draw_objs), 1, 'fileviewer redraw')
    
    # test new
    b.draw_style.set('new')
    
    tab.runn.set(40123)
    tab.get_data()
    tab.entry_asym_type.current(0) # set combined asym mode
    
    tab.draw('inspect')
    tab.draw('inspect')
    
    test_perfect(len(b.plt.plots['inspect']), 3, 'fileviewer draw new')
    
    b.do_close_all()
    
def test_autocomplete():
    tab.year.set(2020)
    tab.runn.set(402)
    tab.get_data()
    test_perfect(tab.data.run, 40299, 'fileviewer autocomplete fetch')
    




# test inspect tab
# Derek Fujimoto
# Feb 2021

from bfit.test.testing import *
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import bdata as bd
from bfit import pulsed_exp, minuit
from bfit.fitting.leastsquares import LeastSquares

# get bfit object and tab
tab = b.fit_files
tab2 = b.fetch_files
b.notebook.select(2)

def get_data():
    
    # clear setup
    b.do_close_all()
    tab2.remove_all()
    
    # fetch
    tab2.year.set(2020)
    tab2.run.set('40123 40127')
    tab2.get_data()

def test_populate():

    get_data()
    
    # check populate
    tab.populate()
    test_misc('2020.40123' in tab.fit_lines.keys(), 'First fetched run populated')
    test_misc('2020.40127' in tab.fit_lines.keys(), 'Second fetched run populated')
    
    # remove
    tab2.data_lines['2020.40123'].degrid()
    tab.populate()
    test_misc('2020.40123' not in tab.fit_lines.keys(), 'First fetched run removed')
    
    # fetch again
    tab2.get_data()
    tab.populate()
    test_misc('2020.40123' in tab.fit_lines.keys(), 'Re-populated after run removal and refetch')
    
def test_populate_param():
        
    get_data()
    tab.populate()
        
    # get fit line
    line = tab.fit_lines['2020.40123']
    
    # set fit fn
    tab.fit_function_title.set('Str Exp')
    tab.populate_param()
    test_perfect(line.get_new_parameters(), ('1_T1', 'amp', 'beta'), 'Single function parameters populated')
    
    # multiple terms
    tab.fit_function_title.set('Exp')
    tab.n_component.set(2)
    tab.populate_param()
    test_perfect(line.get_new_parameters(), ('1_T1_0', '1_T1_1', 'amp_0', 'amp_1'), 
                 'Two term function parameters populated')
    
# do the fit separately
def separate_curve_fit(entry, run, **kwargs):
    
    # data
    year, run = tuple(map(int, run.split('.')))
    dat = bd.bdata(run, year)
    
    # fit function
    pexp = pulsed_exp(lifetime = bd.life.Li8, pulse_len = dat.get_pulse_s())

    p0 = [float(entry[k]['p0'][1].get()) for k in entry.keys()]
    blo = [float(entry[k]['blo'][1].get()) for k in entry.keys()]
    bhi = [float(entry[k]['bhi'][1].get()) for k in entry.keys()]
    
    # fit
    t,a,da = dat.asym('c')
    par, cov = curve_fit(pexp, t, a, sigma=da, absolute_sigma=True, p0=p0, 
                         bounds=[blo, bhi], **kwargs)
    std = np.diag(cov)**0.5
    
    # chi2
    chi = np.sum(((a-pexp(t, *par))/da)**2)/(len(a)-2)
    
    return (par, std, std, chi)

def separate_migrad(entry, run, do_minos=False, **kwargs):
    
    # data
    year, run = tuple(map(int, run.split('.')))
    dat = bd.bdata(run, year)
    
    # fit function
    pexp = pulsed_exp(lifetime = bd.life.Li8, pulse_len = dat.get_pulse_s())

    # get p0, bounds
    p0 = {k.replace('1_T1','lambda_s'):float(entry[k]['p0'][1].get()) for k in entry.keys()}
    bounds = [( float(entry[k]['blo'][1].get()), 
                float(entry[k]['bhi'][1].get())) for k in entry.keys()]
    
    # fit
    t,a,da = dat.asym('c')
    m = minuit(pexp, t, a, da, **p0, limit = bounds, **kwargs)
    m.migrad()
    
    if do_minos: m.minos()
    
    par = m.values
    
    if do_minos: 
        n = len(par)
        lower = np.abs(np.array([m.merrors[i].lower for i in range(n)]))
        upper = np.array([m.merrors[i].upper for i in range(n)])
    else:
        lower = m.errors
        upper = lower
        
    chi = m.fval
        
    return (par, lower, upper, chi)

def separate_minos(entry, run, **kwargs):
    return separate_migrad(entry, run, do_minos=True, **kwargs)
    
def test_fit(separate_fit, minimizer):
    
    # get data
    get_data()
    tab.populate()
    
    # set minimizer
    b.minimizer.set('bfit.fitting.fitter_%s' % minimizer)
    b.set_fit_routine()
    
    # do fit in bfit
    tab.do_fit()
    
    # check return values -----------------------------------------------------
    
    for k, line in tab.fit_lines.items():
    
        entry = line.parentry
        
        # get results separately
        par, lower, upper, chi2 = separate_fit(entry, k)    
        
        # get results displayed
        res = [float(entry[k]['res'][1].get()) for k in entry.keys()]
        low = [float(entry[k]['dres-'][1].get()) for k in entry.keys()]
        upp = [float(entry[k]['dres+'][1].get()) for k in entry.keys()]
        
        # check
        test_arr(par, res, 'Copying %s minimizer results for %s' % (minimizer, k), tol=1e-3)
        test_arr(lower, low, 'Copying %s minimizer lower errors for %s' % (minimizer, k), tol=1e-3)
        test_arr(upper, upp, 'Copying %s minimizer upper errors for %s' % (minimizer, k), tol=1e-3)
        
        # check chi2
        chi = float(entry['1_T1']['chi'][1].get())
        test(chi, chi2, 'Copying chi2 for %s minimizer for %s' % (minimizer, k), tol=1e-2)
    
    # reset minimizer
    b.minimizer.set('bfit.fitting.fitter_curve_fit')
    b.set_fit_routine()

def test_fit_single(separate_fit, minimizer):
    
    # clear setup
    b.do_close_all()
    tab2.remove_all()
    
    # fetch
    tab2.year.set(2020)
    tab2.run.set('40123')
    tab2.get_data()

    tab.populate()
    line = tab.fit_lines['2020.40123']
    entry = line.parentry
    
    # set minimizer
    b.minimizer.set('bfit.fitting.fitter_%s' % minimizer)
    b.set_fit_routine()
    
    # check return value ------------------------------------------------------
    
    # get results separately
    par, lower, upper, chi2 = separate_fit(entry, '2020.40123')    
    tab.do_fit()
    
    # get results displayed
    res = [float(entry[k]['res'][1].get()) for k in entry.keys()]
    low = [float(entry[k]['dres-'][1].get()) for k in entry.keys()]
    upp = [float(entry[k]['dres+'][1].get()) for k in entry.keys()]
    
    # check
    test_arr(par, res, 'Copying %s minimizer results for single run' % minimizer, tol=1e-3)
    test_arr(lower, low, 'Copying %s minimizer lower errors for single run' % minimizer, tol=1e-3)
    test_arr(upper, upp, 'Copying %s minimizer upper errors for single run' % minimizer, tol=1e-3)
    
    # check chi2
    chi = float(entry['1_T1']['chi'][1].get())
    test(chi, chi2, 'Copying chi2 for %s minimizer for single run' % minimizer, tol=1e-2)
    
    # reset minimizer
    b.minimizer.set('bfit.fitting.fitter_curve_fit')
    b.set_fit_routine()

def test_fit_input():
        
    # get data
    get_data()
    
    # set rebin
    tab2.data_lines['2020.40123'].rebin.set(10)
    tab2.data_lines['2020.40127'].rebin.set(2)
    tab.use_rebin.set(True)
    
    # set function
    tab.fit_function_title.set('Bi Exp')
    
    # set ncomp
    tab.n_component.set(2)
    
    # fit
    tab.populate()    
    tab.do_fit()
    
    # fit input: fn_name, ncomp, data_list
    # data_list: [bdfit, pdict, doptions]
    
    # check function
    test_perfect(tab.fit_input[0], 'Bi Exp', "Function name passing to fitter")
    
    # check ncomp
    test_perfect(tab.fit_input[1], 2, "Number of components passing to fitter")
    
    # check
    test_perfect(tab.fit_input[2][0][2]['rebin'], 10, "Rebin passing for file 1")
    test_perfect(tab.fit_input[2][1][2]['rebin'], 2, "Rebin passing for file 2")

def test_fixed():
    
    # get data
    get_data()
    tab.populate()
    line = tab.fit_lines['2020.40123']
    entry = line.parentry
    
    # fixed input -----------------------------------------------------------
    entry['1_T1']['p0'][0].set('1')
    entry['1_T1']['fixed'][0].set(True)
    tab.do_fit()
    
    test(float(entry['1_T1']['res'][1].get()), 1, 'Fixed result')
    test_misc(np.isnan(float(entry['1_T1']['dres-'][1].get())), 'Fixed lower error')
    test_misc(np.isnan(float(entry['1_T1']['dres+'][1].get())), 'Fixed upper error')
    test_misc(float(entry['amp']['res'][1].get()) != 1, 'Unfixed result')
    test_misc(not np.isnan(float(entry['amp']['dres-'][1].get())), 'Unfixed lower error')
    test_misc(not np.isnan(float(entry['amp']['dres+'][1].get())), 'Unfixed upper error')
    
    # unfix
    entry['1_T1']['fixed'][0].set(False)
   
def test_shared():
    
    # get data
    get_data()
    tab.populate()
    line = tab.fit_lines['2020.40123']
    line2 = tab.fit_lines['2020.40127']
    entry = line.parentry
    entry2 = line2.parentry
    
    # shared input
    entry['1_T1']['shared'][0].set(True)
    tab.do_fit()
    
    T11 = float(entry['1_T1']['res'][1].get())
    T12 = float(entry2['1_T1']['res'][1].get())
    
    test(T11, T12, 'Shared result equal')
    
    T11 = float(entry['1_T1']['dres-'][1].get())
    T12 = float(entry2['1_T1']['dres-'][1].get())
    
    test(T11, T12, 'Shared lower error equal')
    
    T11 = float(entry['1_T1']['dres+'][1].get())
    T12 = float(entry2['1_T1']['dres+'][1].get())
    
    test(T11, T12, 'Shared upper error equal')
    
    # check the unshared result
    amp1 = float(entry['amp']['res'][1].get())
    amp2 = float(entry2['amp']['res'][1].get())
    
    test_misc(amp1 != amp2, 'Unshared result not equal')
    
    amp1 = float(entry['amp']['dres-'][1].get())
    amp2 = float(entry2['amp']['dres-'][1].get())
    
    test_misc(amp1 != amp2, 'Unshared lower error not equal')
    
    amp1 = float(entry['amp']['dres+'][1].get())
    amp2 = float(entry2['amp']['dres+'][1].get())
    
    test_misc(amp1 != amp2, 'Unshared upper error not equal')
    
    # unshare
    entry['1_T1']['shared'][0].set(False)
    
def test_modify_for_all_reset_p0():
    
    # get data
    get_data()
    tab.populate()
    line = tab.fit_lines['2020.40123']
    line2 = tab.fit_lines['2020.40127']
    entry = line.parentry
    entry2 = line2.parentry
    
    # modify
    tab.set_as_group.set(True)
    
    initial = {}
    for k in entry.keys():    
        initial[k] = {}
        for c in ('p0', 'blo', 'bhi'):
            
            initial[k][c] = entry[k][c][0].get()
            entry[k][c][0].set(c)
            test_perfect(entry2[k][c][0].get(), c, "Modify all for %s (%s)" % (k,c))
     
    # reset
    tab.do_reset_initial()
    
    for k in entry.keys():    
        for c in ('p0', 'blo', 'bhi'):
            test_perfect(entry[k][c][0].get(), initial[k][c], "Reset p0 for %s (%s)" % (k,c))
                
def test_p0_prior():
    
    # get data
    get_data()
    tab.populate()
    
    # fit
    tab.do_fit()
    b.do_close_all()
    
    # set next p0
    tab.set_prior_p0.set(True)
    
    # get more data
    tab2.run.set('40124')
    tab2.get_data()
    tab.populate()
    
    # get results and compare
    result = tab.fit_lines['2020.40127'].parentry
    p0 = tab.fit_lines['2020.40124'].parentry
    
    for k in result.keys():
        test(float(result[k]['res'][0].get()), float(p0[k]['p0'][0].get()), 
             'Set last result as p0 for new run for %s'%k, tol=1e-5)
    
def test_result_as_p0():
    
    # get data
    get_data()
    tab.populate()
    
    # fit
    tab.do_fit()
    b.do_close_all()
    
    # set result as P0
    tab.do_set_result_as_initial()
    
    # check
    for run_id in tab.fit_lines.keys():
        entry = tab.fit_lines[run_id].parentry
    
        for k in entry.keys():
            test(float(entry[k]['res'][0].get()), float(entry[k]['p0'][0].get()), 
                'Set result as p0 for %s'%k, tol=1e-5)
        
def test_draw_fit_results():
    pass





# test least squares object
# Derek Fujimoto
# Feb 2021

from bfit.test.testing import *
from bfit.fitting.leastsquares import LeastSquares
import numpy as np

# set up data to test leastsquares object
fn = lambda x, a, b : a*x+b
x = np.arange(0,2)
y = x+1
dy = [1, 2]
dx = [1, 4]
dxl = [1, 8]
dyl = [1, 10]

def test_no_errors():
    ls = LeastSquares(fn, x, y)
    test(0, ls((1,1)), "least squares no errors good parameters")
    test(2, ls((1,2)), "least squares no errors bad parameters")
    
def test_dy():
    ls = LeastSquares(fn, x, y, dy)
    test(0, ls((1,1)), "least squares dy good parameters")
    test(1/4, ls((2,1)), "least squares dy bad parameters")

def test_dx():
    ls = LeastSquares(fn, x, y, dx=dx)
    test(0, ls((1,1)), "least squares dx good parameters")
    test(1/64, ls((2,1)), "least squares dx bad parameters")

def test_dxdy():
    ls = LeastSquares(fn, x, y, dy, dx)
    test(0, ls((1,1)), "least squares dxdy good parameters")
    test(1/(64+4), ls((2,1)), "least squares dxdy bad parameters")

def test_dya():
    ls = LeastSquares(fn, x, y, dy, dy_low=dyl)
    test(0, ls((1,1)), "least squares dy asymmetric good parameters")
    test(1/100, ls((0,1)), "least squares dy asymmetric low parameters")
    test(1/4, ls((2,1)), "least squares dy asymmetric high parameters")

def test_dxa():
    ls = LeastSquares(fn, x, y, dx=dx, dx_low=dxl)
    test(0, ls((1,1)), "least squares dx asymmetric good parameters")
    test(1/144, ls((2,1)), "least squares dx asymmetric bad parameters")

def test_dx_dya():
    ls = LeastSquares(fn, x, y, dy=dy, dx=dx, dy_low=dyl)
    test(0, ls((1,1)), "least squares dx and dy asymmetric good parameters")
    test(4/(100+16), ls((-1,1)), "least squares dx and dy asymmetric low parameters")
    test(1/(4+64), ls((2,1)), "least squares dx and dy asymmetric high parameters")
    
def test_dxa_dy():
    ls = LeastSquares(fn, x, y, dy=dy, dx=dx, dx_low=dxl)
    test(0, ls((1,1)), "least squares dx and dy asymmetric good parameters")
    test(1/(4+144), ls((2,1)), "least squares dy and dx asymmetric bad parameters")
    
def test_dxa_dya():
    ls = LeastSquares(fn, x, y, dy=dy, dx=dx, dy_low=dyl, dx_low=dxl)
    test(0, ls((1,1)), "least squares dx asymmetric and dy asymmetric good parameters")
    test(4/(100+36), ls((-1,1)), "least squares dx asymmetric and dy asymmetric low parameters")
    test(1/(4+144), ls((2,1)), "least squares dx asymmetric and dy asymmetric high parameters")



# test minuit object
# Derek Fujimoto
# Feb 2021

from bfit.test.testing import *
from bfit.fitting.minuit import minuit
import numpy as np

# test inputs
fn = lambda x, a, b, c, d, e: a+b+c+d+e+x
fn2 = lambda x, par: a+b+c+d+e+x
x = np.arange(10)
y = x**2

def test_start():
    m = minuit(fn, x, y, start=[1,2,3,4,5])
    test_perfect(m.values['e'], 5, 'minuit start assignment named ')
    
    m = minuit(fn2, x, y, start=[1,2,3,4,5])
    test_perfect(m.values['x4'], 5, 'minuit start assignment unnamed ')
    
    m = minuit(fn, x, y, e=5)
    test_perfect(m.values['e'], 5, 'minuit start assignment single named ')
    
    m = minuit(fn, x, y)
    test_perfect(m.values['e'], 1, 'minuit start default param named')
    
def test_name():
    m = minuit(fn2, x, y, name=['a', 'b', 'c', 'd', 'e'])
    test_perfect(m.values['e'], 1, 'minuit start default param unnamed ')
    test_perfect(m.values['e'], 1, 'minuit name assignment')
    
def test_error():
    m = minuit(fn, x, y, error=[1,2,3,4,5])
    test_perfect(m.errors['e'], 5, 'minuit errors assignment')
    
    m = minuit(fn, x, y, error=5)
    test_perfect(m.errors['e'], 5, 'minuit errors broadcasting')
    
    m = minuit(fn, x, y, error_e=5)
    test_perfect(m.errors['e'], 5, 'minuit errors name single assignment')
    
def test_limit():
    m = minuit(fn, x, y, limit=[[0,1],[0,2],[0,3],[0,4],[0,5]])
    test_perfect(m.limits['e'][1], 5, 'minuit limits assignment')    
    
    m = minuit(fn, x, y, limit=[0,5])
    test_perfect(m.limits['e'][1], 5, 'minuit limits broadcasting')    
    
def test_fix():
    m = minuit(fn, x, y, fix=[True, True, True, True, True])
    test_perfect(m.fixed['e'], True, 'minuit fixed assignment')
    
    m = minuit(fn, x, y, fix=True)
    test_perfect(m.fixed['e'], True, 'minuit fixed broadcasting')
    
    



# functions to help test
# Derek Fujimoto
# Feb 2021

from bfit.gui.bfit import bfit
import numpy as np

def test(a, b, msg, tol=1e-9):
    if abs(a-b) < tol:
        print("Success: Tested %s accurate to a tolerance of %g" % (msg, tol))
    else:
        raise AssertionError(msg + ": " + str(a) + ' != ' + str(b))    

def test_perfect(a, b, msg):
    if a == b:
        print("Success: Tested %s" % msg)
    else:
        raise AssertionError(msg + ": " + str(a) + ' != ' + str(b))    

def test_misc(test, msg):
    if test:
        print("Success: Tested %s" % msg)
    else:
        raise AssertionError(msg + ": " + str(test) + " returns False")    

def test_arr(a, b, msg, tol=1e-9):
    if all(abs(np.array(a)-np.array(b)) < tol):
        print("Success: Tested %s accurate to a tolerance of %g" % (msg, tol))
    else:
        print(a)
        print(b)
        raise AssertionError(msg + " failed comparison")

def test_action(action, msg, *args, **kwargs):
    try:
        action(*args, **kwargs)
    except Exception as err:
        raise AssertionError("Failure: %s with error %s" % (msg, err)) from None
    else:
        print("Success: %s " % msg)
    
# make gui
b = bfit(None, True)



# Model the fit results with a function
# Derek Fujimoto
# August 2019

from tkinter import *
from tkinter import ttk, messagebox
from functools import partial
import logging
import numpy as np

from bfit import logger_name
from bfit.gui.template_fit_popup import template_fit_popup
from bfit.backend.raise_window import raise_window
from bfit.fitting.minuit import minuit

# ========================================================================== #
class popup_fit_results(template_fit_popup):
    """
        Popup window for modelling the fit results with a function
        
        chi_label:      Label, chisquared output
        fittab:         notebook tab
        par:            list, fit results
        reserved_pars:  dict, keys: x, y vals: strings of parameter names
        
        text:           string, model text
        
        xaxis:          StringVar, x axis drawing/fitting parameter
        yaxis:          StringVar, y axis drawing/fitting parameter
        
    """

    # names of modules the constraints have access to
    modules = {'np':'numpy'}

    window_title = 'Fit the results with a model'
    reserved_pars = ['x', 'y']

    # ====================================================================== #
    def __init__(self, bfit, input_fn_text='', output_par_text='', output_text='', 
                 chi=np.nan, x='', y=''):
        
        super().__init__(bfit, input_fn_text, output_par_text, output_text)
        self.fittab = self.bfit.fit_files
        self.chi = chi
        
        # menus for x and y values
        axis_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        
        ttk.Label(  axis_frame, 
                    text='Variable definitions:\n', 
                    justify=LEFT).grid(column=0, row=0, columnspan=2, sticky=W)
        ttk.Label(axis_frame, text="x axis:").grid(column=0, row=1)
        ttk.Label(axis_frame, text="y axis:").grid(column=0, row=2)
        ttk.Label(axis_frame, text=' ').grid(column=0, row=3)
        
        self.xaxis = StringVar()
        self.yaxis = StringVar()
        
        if x:   self.xaxis.set(x)
        else:   self.xaxis.set(self.fittab.xaxis.get())
        
        if y:   self.yaxis.set(y)
        else:   self.yaxis.set(self.fittab.yaxis.get())
        
        self.xaxis_combobox = ttk.Combobox(axis_frame, textvariable=self.xaxis, 
                                      state='readonly', width=19)
        self.yaxis_combobox = ttk.Combobox(axis_frame, textvariable=self.yaxis, 
                                      state='readonly', width=19)
        
        self.xaxis_combobox['values'] = self.fittab.xaxis_combobox['values']
        self.yaxis_combobox['values'] = self.fittab.yaxis_combobox['values']
        
        # module names 
        module_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        s = 'Reserved module names:\n\n'
        
        maxk = max(list(map(len, list(self.modules.keys()))))
        
        s += '\n'.join(['%s:   %s' % (k.rjust(maxk), d) for k, d in self.modules.items()])
        s += '\n'
        modules_label = ttk.Label(module_frame, text=s, justify=LEFT)
        
        # chisquared output
        chi_frame = ttk.Frame(self.left_frame, relief='sunken', pad=5)
        self.chi_label = ttk.Label(chi_frame, 
                                    text='ChiSq: %.2f' % np.around(chi, 2), 
                                    justify=LEFT)
        
        # Text entry
        self.entry_label['text'] = 'Enter a one line equation using "x"'+\
                                 ' to model y(x)'+\
                                 '\nEx: "y = a*x+b"'
             
        # draw button
        button_draw = ttk.Button(self.right_frame, text='Draw', 
                                 command=self.draw_model, pad=1)
                
        # gridding
        modules_label.grid(column=0, row=0)
        self.chi_label.grid(column=0, row=0)
        self.xaxis_combobox.grid(column=1, row=1)
        self.yaxis_combobox.grid(column=1, row=2)
        
        axis_frame.grid(column=0, row=0, rowspan=1, sticky=(E, W), padx=1, pady=1)
        module_frame.grid(column=0, row=1, sticky=(E, W), padx=1, pady=1)
        chi_frame.grid(column=0, row=2, sticky=(E, W), padx=1, pady=1, rowspan=2)
        
        button_draw.grid(column=0, row=3, sticky=(E, W))
        
    # ====================================================================== #
    def _do_fit(self, text):
        
        # save model text
        self.text= text[-1]
        
        # get fit data
        xstr = self.xaxis.get()
        ystr = self.yaxis.get()
        
        # Make model
        parnames = self.output_par_text.get('1.0', END).split('\n')[:-1]
        parstr = ', '.join(parnames)
        eqn = text[-1].split('=')[-1]
        model = 'lambda x, %s : %s' % (parstr, eqn)
        
        self.logger.info('Fitting model %s for x="%s", y="%s"', model, xstr, ystr)
        
        model = eval(model)
        self.model_fn = model
        npar = len(parnames)
        
        # set up p0, bounds
        p0 = self.new_par['p0'].values
        blo = self.new_par['blo'].values
        bhi = self.new_par['bhi'].values
        
        p0 = list(map(float, p0))
        blo = list(map(float, blo))
        bhi = list(map(float, bhi))
                    
        # get data to fit
        xvals, xerrs_l, xerrs_h = self._get_data(xstr)
        yvals, yerrs_l, yerrs_h = self._get_data(ystr)
        
        # minimize
        m = minuit(model, xvals, yvals, 
                  dy = yerrs_h, 
                  dx = xerrs_h, 
                  dy_low = yerrs_l, 
                  dx_low = xerrs_l,
                  name = parnames,
                  print_level = 0,
                  limit = np.array([blo, bhi]).T,
                  )
        
        m.migrad()
        m.hesse()
        m.minos()
        
        # print fitting quality
        try:
            print(m.fmin)
            print(m.params)
            print(m.merrors)
        except UnicodeEncodeError:
            pass
        
        # get results
        par = m.values
        self.par = par
        n = len(m.merrors)
        
        if npar == 1:
            std_l = np.array([m.merrors[i].lower for i in range(n)])
            std_h = np.array([std_l])
        else:
            std_l = np.array([m.merrors[i].lower for i in range(n)])
            std_h = np.array([m.merrors[i].upper for i in range(n)])
            
        # chi2
        chi = m.chi2()
        self.chi_label['text'] = 'ChiSq: %.2f' % np.around(chi, 2)
        self.chi = chi
        
        self.logger.info('Fit model results: %s, Errors-: %s, Errors+: %s', 
                        str(par), str(std_l), str(std_h))
        
        self.bfit.plt.figure('param')
        self.draw_data()
        self.draw_model()    
        
        return (par, std_l, std_h)
        
    # ====================================================================== #
    def _get_data(self, xstr):
        """
            Get data and clean
            
            xstr: label for data to fetch
        """
        
        # get data
        try:
            xvals, xerrs = self.fittab.get_values(xstr)
        except UnboundLocalError as err:
            self.logger.error('Bad input parameter')
            messagebox.showerror("Error", 'Bad input parameter')
            raise err
        except (KeyError, AttributeError) as err:
            self.logger.error('Parameter "%s not found for fitting', xstr)
            messagebox.showerror("Error", 'Parameter "%s" not found' % xstr)
            raise err
        
        # split errors    
        if type(xerrs) is tuple: 
            xerrs_l = xerrs[0]
            xerrs_h = xerrs[1]
        else:
            xerrs_l = xerrs
            xerrs_h = xerrs
        
        xvals = np.asarray(xvals)
        xerrs_l = np.asarray(xerrs_l)
        xerrs_h = np.asarray(xerrs_h)
                        
        # check errors 
        if all(np.isnan(xerrs_l)): xerrs_l = None
        if all(np.isnan(xerrs_h)): xerrs_h = None
        
        return (xvals, xerrs_l, xerrs_h)
    
    # ======================================================================= #
    def draw_data(self):
        figstyle = 'param'
        
        # get draw components
        xstr = self.xaxis.get()
        ystr = self.yaxis.get()
        id = self.fittab.par_label.get()
        
        self.logger.info('Draw model fit data "%s" vs "%s"', ystr, xstr)

        # get data
        xvals, xerrs_l, xerrs_h = self._get_data(xstr)
        yvals, yerrs_l, yerrs_h = self._get_data(ystr)

        # sort by x values, check for empty arrays
        idx = np.argsort(xvals)
        xvals = np.asarray(xvals)[idx]
        yvals = np.asarray(yvals)[idx]
        
        if xerrs_h is not None:     xerrs_h = np.asarray(xerrs_h)[idx]
        if xerrs_l is not None:     xerrs_l = np.asarray(xerrs_l)[idx]
        if yerrs_h is not None:     yerrs_h = np.asarray(yerrs_h)[idx]
        if yerrs_l is not None:     yerrs_l = np.asarray(yerrs_l)[idx]
        
        if xerrs_h is None and xerrs_l is None:     xerrs = None
        else:                                       xerrs = (xerrs_l, xerrs_h)
        if yerrs_h is None and yerrs_l is None:     yerrs = None
        else:                                       yerrs = (yerrs_l, yerrs_h)
        
        # get mouseover annotation labels
        mouse_label, _ = self.fittab.get_values('Unique Id')
        mouse_label = np.asarray(mouse_label)[idx]

        #draw data
        self.fittab.plt.errorbar(figstyle, id, xvals, yvals, 
                                 yerr=yerrs, 
                                 xerr=xerrs, 
                                 fmt='.', 
                                 annot_label=mouse_label)

        # pretty labels
        xstr = self.fittab.fitter.pretty_param.get(xstr, xstr)
        ystr = self.fittab.fitter.pretty_param.get(ystr, ystr)
        
        # plot elements
        self.fittab.plt.xlabel(figstyle, xstr)
        self.fittab.plt.ylabel(figstyle, ystr)
        self.fittab.plt.tight_layout(figstyle)

        raise_window()

    # ======================================================================= #
    def draw_model(self):
        figstyle = 'param'
        
        self.logger.info('Draw model "%s"', self.text)
        
        # get fit function and label id
        fn = self.model_fn
        id = self.fittab.par_label.get()
        
        # get x data
        xstr = self.xaxis.get()
        xvals, _, _ = self._get_data(xstr)
        
        # draw fit
        fitx = np.linspace(min(xvals), max(xvals), self.fittab.n_fitx_pts)
        f = self.fittab.plt.plot(figstyle, id+self.text, fitx, fn(fitx, *self.par), 
                                 color='k', label=self.text)
                
        raise_window()



import bfit.backend.colors as colors

# =========================================================================== #
def on_entry_click(event, entry, text):
    """Vanish grey text on click"""
    if entry.get() == text:
        entry.delete(0, "end") # delete all the text in the entry
        entry.insert(0, '') #Insert blank for user input
        entry.config(foreground = colors.entry_white)

# =========================================================================== #
def on_focusout(event, entry, text):
    """Set grey text for boxes on exit"""
    if entry.get() == '':
        entry.insert(0, text)
        entry.config(foreground = colors.entry_grey)
    else:
        entry.config(foreground = colors.entry_white)



# color definitions
foreground          = 'white'
background          = 'black'
fieldbackground     = 'grey30'
focusbackground     = 'grey'
insertbackground    = 'white'

selected            = 'green2'
    
readonly            = 'grey10'
disabled            = 'grey40'
    
entry_grey          = 'grey45'
entry_white         = 'grey92'
entry_disabled      = 'black'
    
tab                 = 'grey'
    
red                 = 'red3'
    
font                = 'TkFixedFont'



# Data object for holding bdata and related file settings for drawing and 
# fitting. 
# Derek Fujimoto
# Nov 2018

from tkinter import *
from bdata import bdata, bmerged
from bfit.gui.calculator_nqr_B0 import current2field
from bfit import logger_name

import numpy as np
import pandas as pd

import bfit
import logging
import textwrap

# =========================================================================== #
# =========================================================================== #
class fitdata(object):
    """
        Hold bdata and related file settings for drawing and fitting in fetch 
        files tab and fit files tab. 
        
        Data Fields:
            
            bd:         bdata object for data and asymmetry (bdata)
            bfit:       pointer to top level parent object (bfit)
            bias:       platform bias in kV (float)
            bias_std:   platform bias in kV (float)
            check_state:(BooleanVar)  
            chi:        chisquared from fit (float)
            deadtime:   deadtime value
            drawarg:    drawing arguments for errorbars (dict)
            field:      magnetic field in T (float)
            field_std:  magnetic field standard deviation in T (float)
            fitfn:      function (function pointer)
            fitfnname:  function (str)
            fitpar:     initial parameters {column:{parname:float}} and results
                        Columns are fit_files.fitinputtab.collist
            id:         key for unique idenfication (str)    
            label:      label for drawing (StringVar)
            mode:       run mode (str)
            omit:       omit bins, 1f only (StringVar)
            parnames:   parameter names in the order needed by the fit function
            rebin:      rebin factor (IntVar)
            run:        run number (int)
            year:       run year (int)
              
    """
     
    # ======================================================================= #
    def __init__(self, parentbfit, bd):
        
        # get logger
        self.logger = logging.getLogger(logger_name)
        self.logger.debug('Initializing run %d (%d).', bd.run, bd.year)
        
        # top level pointer
        self.bfit = parentbfit
        
        # bdata access
        self.bd = bd
        
        # input variables for tkinter
        self.rebin = IntVar()
        self.omit = StringVar()
        self.label = StringVar()
        self.check_state = BooleanVar()
        
        self.check_state.set(False)
        
        # key for IDing file 
        self.id = self.bfit.get_run_key(data=bd)
        
        # initialize fitpar with fitinputtab.collist
        self.fitpar = pd.DataFrame([], columns=['p0', 'blo', 'bhi', 'res', 
                                    'dres+', 'dres-', 'chi', 'fixed', 'shared'])
        self.read()

    # ======================================================================= #
    def __getattr__(self, name):
        """Access bdata attributes in the case that fitdata doesn't have it."""
        try:
            return self.__dict__[name]
        except KeyError:
            return getattr(self.bd, name)

    # ======================================================================= #
    def asym(self, *args, **kwargs):
        
        deadtime = 0
        
        # check if deadtime corrections are needed
        if self.bfit.deadtime_switch.get():
            
            # check if corrections should be calculated for each run
            if self.bfit.deadtime_global.get():
                deadtime = self.bfit.deadtime
            else:
                deadtime = self.bd.get_deadtime(c=self.bfit.deadtime, fixed='c')
                
        return self.bd.asym(*args, deadtime=deadtime, **kwargs)

    # ======================================================================= #
    @property
    def beam_kev(self): 
        try:
            return self.bd.beam_keV
        except AttributeError:
            return np.nan
    
    @property
    def beam_kev_err(self): 
        try:
            return self.bd.beam_keV_err
        except AttributeError:
            return np.nan
        
    # ======================================================================= #
    def get_temperature(self, channel='A'):
        """
            Get the temperature of the run.
            Return (T, std T)
        """
        
        try:
            if channel == 'A':
                T = self.bd.camp['smpl_read_A'].mean
                dT = self.bd.camp['smpl_read_A'].std
            elif channel == 'B':
                T = self.bd.camp['smpl_read_B'].mean
                dT = self.bd.camp['smpl_read_B'].std
            elif channel == '(A+B)/2':
                Ta = self.bd.camp['smpl_read_A'].mean
                Tb = self.bd.camp['smpl_read_B'].mean
                dTa = self.bd.camp['smpl_read_A'].std
                dTb = self.bd.camp['smpl_read_B'].std
                
                T = (Ta+Tb)/2
                dT = ((dTa**2+dTb**2)**0.5)/2
            else:
                raise AttributeError("Missing required temperature channel.")
        
        except KeyError:
            T = np.nan
            dT = np.nan
        
        return (T, dT)
        
    # ======================================================================= #
    def read(self):
        """Read data file"""
        
        # bdata access
        if type(self.bd) is bdata:
            self.bd = bdata(self.run, self.year)
        elif type(self.bd) is bmerged:
            years = list(map(int, textwrap.wrap(str(self.year), 4)))
            runs = list(map(int, textwrap.wrap(str(self.run), 5)))
            self.bd = bmerged([bdata(r, y) for r, y in zip(runs, years)])
                
        # set temperature 
        try:
            self.temperature = temperature_class(*self.get_temperature(self.bfit.thermo_channel.get()))
        except AttributeError as err:
            self.logger.exception(err)
            try:
                self.temperature = self.bd.camp.oven_readC
            except AttributeError:
                self.logger.exception('Thermometer oven_readC not found')
                self.temperature = -1111
        
        # field
        try:
            if self.bd.area == 'BNMR':
                self.field = self.bd.camp.b_field.mean
                self.field_std = self.bd.camp.b_field.std
            else:
                self.field = current2field(self.bd.epics.hh_current.mean)*1e-4
                self.field_std = current2field(self.bd.epics.hh_current.std)*1e-4
        except AttributeError:
            self.logger.exception('Field not found')
            self.field = np.nan
            self.field_std = np.nan
            
        # bias
        try:
            if self.bd.area == 'BNMR': 
                self.bias = self.bd.epics.nmr_bias.mean
                self.bias_std = self.bd.epics.nmr_bias.std
            else:
                self.bias = self.bd.epics.nqr_bias.mean/1000.
                self.bias_std = self.bd.epics.nqr_bias.std/1000.
        except AttributeError:
            self.logger.exception('Bias not found')
            self.bias = np.nan
            
    # ======================================================================= #
    def set_fitpar(self, values):
        """Set fitting initial parameters
        values: output of routine gen_init_par: DataFrame:            
                columns: [p0, blo, bhi, fixed]
                index: parameter names
        """
    
        self.parnames = values.index.values
    
        for v in self.parnames:
            for c in values.columns:
                self.fitpar.loc[v, c] = values.loc[v, c]
    
        self.logger.debug('Fit initial parameters set to %s', self.fitpar)

    # ======================================================================= #
    def set_fitresult(self, values):
        """
            Set fit results. Values is output of fitting routine. 
            
            values: {fn: function handle, 
                     'results': DataFrame of fit results,
                     'gchi': global chi2}
                     
            values['results']: 
                columns: [res, dres+, dres-, chi, fixed, shared]
                index: parameter names
        """
        
        # set function
        self.fitfn = values['fn']
        
        # get data frame
        df = values['results']
        
        # set parameter names
        self.parnames = df.index.values
        
        # set chi
        self.chi = df['chi'].values[0]
        
        # set parameters
        for v in self.parnames:
            for c in df.columns:
                self.fitpar.loc[v, c] = df.loc[v, c]
        self.logger.debug('Setting fit results to %s', self.fitpar)
    
# ========================================================================== #
class temperature_class(object):
    """
        Emulate storage container for camp variable smpl_read_%
    """
    
    def __init__(self, mean, std):
        self.mean = mean
        self.std = std
    
